class PowerwallError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class ApiError(PowerwallError):
    def __init__(self, error):
        super().__init__(f"Powerwall api error: {error}")


class APIChangedError(ApiError):
    def __init__(self, response_class, json_response, added_attrs=[], removed_attrs=[]):
        self.json_response = json_response
        self.response_class = response_class
        self.added_attrs = added_attrs
        self.removed_attrs = removed_attrs

        msg = self._construct_msg()

        super().__init__(msg)

    def _construct_msg(self):
        msg = f"It seems like the Powerwall API changed for {self.response_class}"
        if len(self.added_attrs) > 0:
            if len(self.removed_attrs) > 0:
                msg = f"{msg}: Attributes added: {self.added_attrs}, removed attributes {self.removed_attrs}"
            else:
                msg = f"{msg}: Some attributes where added to the response: {self.added_attrs}"
        elif len(self.removed_attrs) > 0:
            msg = f"{msg}: Some attributes where removed from the response: {self.removed_attrs}"
        return msg


class PowerwallUnreachableError(PowerwallError):
    def __init__(self, reason=None):
        msg = f"Powerwall is unreachable"
        self.reason = reason
        if reason is not None:
            msg = f"{msg}: {reason}"
        super().__init__(msg)


class AccessDeniedError(PowerwallError):
    def __init__(self, resource, error=None):
        self.resource = resource
        self.error = error
        msg = f"Access denied for resource {resource}"
        if error is not None:
            msg = f"{msg}: {error}"
        super().__init__(msg)



from json.decoder import JSONDecodeError
from typing import List, Union
from urllib.parse import urljoin, urlparse, urlsplit, urlunparse, urlunsplit

import requests
from requests import Session
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning

from .const import (
    DEFAULT_KW_ROUND_PERSICION,
    SUPPORTED_OPERATION_MODES,
    SUPPORTED_POWERWALL_VERSIONS,
    DeviceType,
    GridState,
    GridStatus,
    LineStatus,
    MeterType,
    OperationMode,
    User,
)
from .error import AccessDeniedError, ApiError, PowerwallUnreachableError
from .helpers import convert_to_kw
from .responses import (
    CustomerRegistrationResponse,
    ListPowerwallsResponse,
    LoginResponse,
    MeterDetailsResponse,
    MetersAggregateResponse,
    MetersResponse,
    PowerwallsStatusResponse,
    PowerwallStatusResponse,
    SiteInfoResponse,
    SitemasterResponse,
    SolarsResponse,
)

VERSION = "0.2.5"


class Powerwall(object):
    def __init__(
        self,
        endpoint: str,
        timeout: int = 10,
        http_session: requests.Session = None,
        verify_ssl: bool = False,
        disable_insecure_warning: bool = True,
        dont_validate_response: bool = True,
    ):

        if disable_insecure_warning:
            disable_warnings(InsecureRequestWarning)

        if endpoint.startswith("https"):
            self._endpoint = endpoint
        elif endpoint.startswith("http"):
            self._endpoint = endpoint.replace("http", "https")
        else:
            self._endpoint = f"https://{endpoint}"

        if not self._endpoint.endswith("api") and not self._endpoint.endswith("/"):
            self._endpoint += "/api/"
        elif self._endpoint.endswith("api"):
            self._endpoint += "/"
        elif self._endpoint.endswith("/") and not self._endpoint.endswith("api/"):
            self._endpoint += "api/"

        self._timeout = timeout
        self._http_session = http_session if http_session else Session()
        self._http_session.verify = verify_ssl
        self._dont_validate_response = dont_validate_response

        self._token = None

    def _process_response(self, response: str) -> dict:
        if response.status_code == 404:
            raise ApiError(f"The url {response.request.path_url} returned error 404")

        if response.status_code == 401 or response.status_code == 403:
            response_json = None
            try:
                response_json = response.json()
            except Exception:
                raise AccessDeniedError(response.request.path_url)
            else:
                raise AccessDeniedError(
                    response.request.path_url, response_json["error"]
                )

        if response.status_code == 502:
            raise PowerwallUnreachableError()

        try:
            response_json = response.json()
        except JSONDecodeError:
            raise ApiError(f"Error while decoding json of response: {response.text}")

        if response_json is None:
            return {}

        if "error" in response_json:
            raise ApiError(response_json["error"])

        return response_json

    def _get(
        self, path: str, needs_authentication: bool = False, headers: dict = {}
    ) -> dict:
        if needs_authentication and not self.is_authenticated():
            raise ApiError(f"Authentication required to access {path}")

        try:
            response = self._http_session.get(
                url=urljoin(self._endpoint, path),
                timeout=self._timeout,
                headers=headers,
            )
        except requests.exceptions.ConnectionError as e:
            raise PowerwallUnreachableError(e)

        return self._process_response(response)

    def _post(
        self,
        path: str,
        payload: dict,
        needs_authentication: bool = False,
        headers: dict = {},
    ) -> dict:
        if needs_authentication and not self.is_authenticated():
            raise ApiError(f"Authentication required to access {path}")

        try:
            response = self._http_session.post(
                url=urljoin(self._endpoint, path),
                data=payload,
                timeout=self._timeout,
                headers=headers,
            )
        except requests.exceptions.ConnectionError as e:
            raise PowerwallUnreachableError(e)

        return self._process_response(response)

    def is_authenticated(self) -> bool:
        return "AuthCookie" in self._http_session.cookies.keys()

    def login_as(
        self,
        user: Union[User, str],
        email: str,
        password: str,
        force_sm_off: bool = False,
    ) -> LoginResponse:
        if isinstance(user, User):
            user = user.value

        # force_sm_off is referred to as 'shouldForceLogin' in the web source code
        response = self._post(
            "login/Basic",
            {
                "username": user,
                "email": email,
                "password": password,
                "force_sm_off": force_sm_off,
            },
        )

        # The api returns an auth cookie which is automatically set
        # so there is no need to further process the response

        return LoginResponse(response, no_check=self._dont_validate_response)

    def login(
        self, email: str, password: str, force_sm_off: bool = False
    ) -> LoginResponse:
        return self.login_as(User.CUSTOMER, email, password, force_sm_off)

    def logout(self):
        if not self.is_authenticated():
            raise ApiError("Must be logged in to log out")
        # The api unsets the auth cookie and the token is invalidated
        self._get("logout", True)

    def run(self):
        self._get("sitemaster/run", True)

    def stop(self):
        self._get("sitemaster/stop", True)

    def get_charge(self, rounded: bool = True) -> Union[float, int]:
        """Returns current charge of powerwall"""
        charge = self._get("system_status/soe")["percentage"]
        if rounded:
            return round(charge)
        else:
            return charge

    def get_sitemaster(self) -> SitemasterResponse:
        return SitemasterResponse(
            self._get("sitemaster"), no_check=self._dont_validate_response
        )

    def get_meters(self) -> MetersAggregateResponse:
        """Returns the different meters in a MetersAggregateResponse"""
        return MetersAggregateResponse(
            self._get("meters/aggregates"), no_check=self._dont_validate_response
        )

    def get_meter_details(
        self, meter: MeterType
    ) -> Union[List[MeterDetailsResponse], None]:
        """Returns details about a specific meter. 
        If their are no details available for a meter None is returned."""
        resp = self._get(f"meters/{meter.value}")
        if isinstance(resp, list):
            return [
                MeterDetailsResponse(item, no_check=self._dont_validate_response)
                for item in resp
            ]
        elif isinstance(resp, dict):
            return None

    def get_meter_readings(self) -> dict:
        return self._get("meter/readings", True)

    def get_grid_status(self) -> GridStatus:
        """Returns the current grid status."""
        return GridStatus(self._get("system_status/grid_status")["grid_status"])

    def get_grid_services_active(self) -> bool:
        return self._get("system_status/grid_status")["grid_services_active"]

    def get_grid_codes(self) -> list:
        """Returns all available grid codes"""
        return self._get("site_info/grid_codes", needs_authentication=True)

    def get_site_info(self) -> SiteInfoResponse:
        """Returns information about the powerwall site"""
        return SiteInfoResponse(
            self._get("site_info"), no_check=self._dont_validate_response
        )

    def set_site_name(self, site_name: str):
        return self._post("site_info/site_name", {"site_name": site_name}, True)

    def get_status(self) -> PowerwallStatusResponse:
        return PowerwallStatusResponse(
            self._get("status"), no_check=self._dont_validate_response
        )

    def get_powerwalls_status(self) -> PowerwallsStatusResponse:
        return PowerwallsStatusResponse(
            self._get("powerwalls/status", True), no_check=self._dont_validate_response
        )

    def get_device_type(self) -> DeviceType:
        """Returns the device type of the powerwall"""
        return DeviceType(self._get("device_type")["device_type"])

    def get_customer_registration(self) -> CustomerRegistrationResponse:
        return CustomerRegistrationResponse(
            self._get("customer/registration"), no_check=self._dont_validate_response
        )

    def get_powerwalls(self) -> ListPowerwallsResponse:
        """Returns powerwalls and status"""
        return ListPowerwallsResponse(
            self._get("powerwalls"), no_check=self._dont_validate_response
        )

    def get_serial_numbers(self) -> List[str]:
        return [
            powerwall.PackageSerialNumber
            for powerwall in self.get_powerwalls().powerwalls
        ]

    def get_operation_mode(self) -> OperationMode:
        return OperationMode(self._get("operation", True)["real_mode"])

    def get_backup_preserve_percentage(self) -> float:
        return self._get("operation", True)["backup_reserve_percentage"]

    # def set_mode_and_backup_preserve_percentage(self, mode, percentage):
    #     self._post("operation", {"mode": mode, "percentage": percentage})

    # def set_backup_preserve_percentage(self, percentage):
    #     self.set_mode_and_backup_preserve_percentage(self.mode, percentage)

    # def set_mode(self, mode):
    #     self.set_mode_and_backup_preserve_percentage(
    #         mode, self.backup_preserve_percentage)

    def get_networks(self) -> list:
        return self._get("networks")

    def get_phase_usage(self):
        return self._get("powerwalls/phase_usages", needs_authentication=True)

    def set_run_for_commissioning(self):
        self._post("sitemaster/run_for_commissioning", True)

    def get_solars(self) -> List[SolarsResponse]:
        solars = self._get("solars", needs_authentication=True)
        return [
            SolarsResponse(solar, no_check=self._dont_validate_response)
            for solar in solars
        ]

    def get_vin(self) -> str:
        return self._get("config", needs_authentication=True)["vin"]

    def get_logs(self) -> str:
        return self._get("getlogs", needs_authentication=True)

    def get_meters_info(self) -> list:
        return self._get("meters", needs_authentication=True)

    def get_installer_info(self) -> dict:
        return self._get("installer", needs_authentication=True)

    def get_solar_brands(self) -> List[str]:
        return self._get("solars/brands", needs_authentication=True)

    def get_version(self) -> str:
        return self.get_status().version

    def get_git_hash(self) -> str:
        return self.get_status().git_hash

    def is_sending_to(self, meter: MeterType, rounded=True) -> bool:
        """Wrapper method for is_sending_to"""
        return self.get_meters().get(meter).is_sending_to()

    def is_drawing_from(self, meter: MeterType) -> bool:
        """Wrapper method for is_drawing_from"""
        return self.get_meters().get(meter).is_drawing_from()

    def is_active(self, meter: MeterType) -> bool:
        """Wrapper method for is_active()"""
        return self.get_meters().get(meter).is_active()

    def get_power(self, meter: MeterType) -> bool:
        return self.get_meters().get(meter).get_power()

    def is_powerwall_version_supported(self) -> bool:
        return self.get_version() in SUPPORTED_POWERWALL_VERSIONS

    def set_dont_validate_response(self, value):
        self._dont_validate_response = value



import re
from datetime import datetime, timedelta

from .const import DEFAULT_KW_ROUND_PERSICION, MeterType
from .error import APIChangedError
from .helpers import convert_to_kw


class Response(object):
    """Basic Response object that can be constructed from a json response"""

    # A list of attributes that should be in the json_response
    _JSON_ATTRS = []
    # A list of attributes that may be in the json_response but aren"t required
    _OPTIONAL_JSON_ATTRS = []

    def __init__(self, json_response: dict, no_check: bool = False):
        self.json_response = json_response
        self._set_attrs(no_check)
        self.response_validated = not no_check

    def _set_attrs(self, no_check: bool = False) -> None:
        """Set attributes from _JSON_ATTRS as object properties. 
        Also checks wether the response is valid. 
        This can be disabled by passing no_check=True"""
        missing_attrs = []
        for attr in self.__class__._JSON_ATTRS:
            self._add_attr(attr, missing_attrs)

        if len(self.__class__._OPTIONAL_JSON_ATTRS) > 0:
            for attr in self.__class__._OPTIONAL_JSON_ATTRS:
                self._add_attr(attr)

        if not no_check:
            added_attrs = self._get_added_attrs()

        # We are missing some attributes in the json_response
        if not no_check and len(missing_attrs) > 0:
            raise APIChangedError(
                self.__class__, self.json_response, added_attrs, missing_attrs
            )

    def _add_attr(self, attr, missing_attrs=[]) -> None:
        # Make sure the attribute also exist in the json_response
        if attr in self.json_response:
            setattr(self, attr, self.json_response[attr])
        else:
            missing_attrs.append(attr)

    def _get_added_attrs(self) -> list:
        added_attrs = []
        for attr in self.json_response.keys():
            if attr not in self.__class__._JSON_ATTRS:
                added_attrs.append(attr)
        return added_attrs

    # Helper methods to make interaction with optional json attributes easier

    def has_optional_attrs(self) -> bool:
        return len(self.__class__._OPTIONAL_JSON_ATTRS) > 0

    def has_optional_attrs_set(self) -> bool:
        """Checks wether all optional attributes are present in the json response"""
        return set(
            self.__class__._OPTIONAL_JSON_ATTRS + self.__class__._JSON_ATTRS
        ) == set(self.json_response.keys())

    def get(self, key, default=None):
        """Equavivalent to dict.get"""
        return self.json_response.get(key, default)

    def has_key(self, key) -> bool:
        return key in self.json_response.keys()

    def __getitem__(self, key):
        return getattr(self, key)

    def __repr__(self) -> str:
        return str(self.json_response)


class MetersResponse(Response):
    """Response for a single Meter
    Usually a nested item in the MetersAggregateResponse
    """

    _JSON_ATTRS = [
        "last_communication_time",
        "instant_power",  # The power that is supplied/drawn from the meter
        "instant_reactive_power",
        "instant_apparent_power",
        "frequency",
        "energy_exported",
        "energy_imported",
        "instant_average_voltage",
        "instant_total_current",
        "i_a_current",
        "i_b_current",
        "i_c_current",
        "timeout",
    ]

    def __init__(self, json_response: dict, meter: MeterType = None, no_check=False):
        super().__init__(json_response, no_check=no_check)
        self.meter = meter

    def is_sending_to(self, precision=DEFAULT_KW_ROUND_PERSICION):
        if self.meter == MeterType.LOAD:
            return convert_to_kw(self.instant_power, precision) > 0
        else:
            return convert_to_kw(self.instant_power, precision) < 0

    def is_drawing_from(self, precision=DEFAULT_KW_ROUND_PERSICION):
        if self.meter == MeterType.LOAD:
            # Cannot draw from load
            return False
        else:
            return convert_to_kw(self.instant_power, precision) > 0

    def is_active(self, precision=DEFAULT_KW_ROUND_PERSICION):
        return convert_to_kw(self.instant_power, precision) != 0

    def get_power(self, precision=DEFAULT_KW_ROUND_PERSICION):
        """Returns power sent/drawn in kWh"""
        return convert_to_kw(self.instant_power, precision)


class MetersAggregateResponse(Response):
    """
    Response for "meters/aggregates"
    """

    def __init__(self, json_response, no_check=False):
        self.json_response = json_response
        for meter in MeterType:
            if meter.value in json_response:
                setattr(
                    self,
                    meter.value,
                    MetersResponse(json_response[meter.value], meter, no_check),
                )

    def get(self, meter: MeterType) -> MetersResponse:
        return getattr(self, meter.value)


class MeterDetailsResponse(Response):
    _JSON_ATTRS = ["id", "location", "type", "cts", "inverted", "connection"]

    def __init__(self, json_response, no_check=False):
        super().__init__(json_response, no_check)
        self.cached_readings = MetersResponse(
            json_response["Cached_readings"], no_check=no_check
        )


class SitemasterResponse(Response):
    _JSON_ATTRS = ["status", "running", "connected_to_tesla"]


class SiteInfoResponse(Response):
    _JSON_ATTRS = [
        "max_site_meter_power_kW",
        "min_site_meter_power_kW",
        "nominal_system_energy_kWh",
        "nominal_system_power_kW",
        "max_system_energy_kWh",
        "max_system_power_kW",
        "site_name",
        "timezone",
        "grid_code",
        "grid_voltage_setting",
        "grid_freq_setting",
        "grid_phase_setting",
        "country",
        "state",
        "region",
    ]

    _OPTIONAL_JSON_ATTRS = ["distributor", "utility", "retailer"]


class CustomerRegistrationResponse(Response):
    _JSON_ATTRS = [
        "privacy_notice",
        "limited_warranty",
        "grid_services",
        "marketing",
        "registered",
        "timed_out_registration",
    ]


class PowerwallStatusResponse(Response):
    _START_TIME_FORMAT = "%Y-%m-%d %H:%M:%S %z"
    _UP_TIME_SECONDS_REGEX = re.compile(
        r"((?P<hours>\d+?)h)((?P<minutes>\d+?)m)((?P<seconds>\d+?).)((?P<microseconds>\d+?)s)"
    )

    _JSON_ATTRS = ["start_time", "up_time_seconds", "is_new", "version", "git_hash"]

    def __init__(self, json_response, no_check=False):
        self.json_response = json_response

        resp = json_response
        resp["start_time"] = datetime.strptime(
            json_response["start_time"], PowerwallStatusResponse._START_TIME_FORMAT
        )
        resp["up_time_seconds"] = self._parse_uptime_seconds(
            json_response["up_time_seconds"]
        )

        self._set_attrs()

    def _parse_uptime_seconds(self, up_time_seconds: str):
        match = self.__class__._UP_TIME_SECONDS_REGEX.match(up_time_seconds)
        if not match:
            raise ValueError(f"Unable to parse up time seconds {up_time_seconds}")

        time_params = {}
        for (name, param) in match.groupdict().items():
            if param:
                time_params[name] = int(param)

        return timedelta(**time_params)


class PowerwallsStatusResponse(Response):
    _JSON_ATTRS = [
        "enumerating",
        "updating",
        "checking_if_offgrid",
        "running_phase_detection",
        "phase_detection_last_error",
        "bubble_shedding",
        "on_grid_check_error",
        "grid_qualifying",
        "grid_code_validating",
        "phase_detection_not_available",
    ]


class ListPowerwallsResponse(Response):
    class PowerwallResponse(Response):
        _JSON_ATTRS = [
            "PackagePartNumber",
            "PackageSerialNumber",
            "type",
            "grid_state",
            "grid_reconnection_time_seconds",
            "under_phase_detection",
            "updating",
            "commissioning_diagnostic",
            "update_diagnostic",
        ]

    _JSON_ATTRS = ["powerwalls", "has_sync", "sync", "states"]

    def __init__(self, json_response, no_check=False):
        super().__init__(json_response, no_check)
        self.status = PowerwallsStatusResponse(self.json_response, no_check=no_check)
        self.powerwalls = [
            ListPowerwallsResponse.PowerwallResponse(powerwall, no_check=no_check)
            for powerwall in self.powerwalls
        ]


class SolarsResponse(Response):
    _JSON_ATTRS = ["brand", "model", "power_rating_watts"]


class LoginResponse(Response):
    _JSON_ATTRS = [
        "email",
        "firstname",
        "lastname",
        "roles",
        "token",
        "provider",
        "loginTime",
    ]
